# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic
import pytest

from canonical import ISO3166Alpha2


class Model(pydantic.BaseModel):
    country: ISO3166Alpha2


@pytest.mark.parametrize("value", [
    "NL",
    "BE",
    "DE"
])
def test_valid_codes(value: str):
    ISO3166Alpha2.validate(value)
    obj = Model.model_validate({'country': value})
    assert isinstance(obj.country, ISO3166Alpha2)


@pytest.mark.parametrize("value", [
    "B1",
    "ABC",
])
def test_invalid_codes(value: str):
    with pytest.raises(ValueError):
        ISO3166Alpha2.validate(value)


def test_eu_greece():
    code = ISO3166Alpha2.validate('EL')
    assert code == 'GR'


def test_eu_northern_ireland():
    code = ISO3166Alpha2.validate('XI')
    assert code == 'XI'


def test_eu_united_kingdom():
    code = ISO3166Alpha2.validate('UK')
    assert code == 'GB'


def test_name_nl():
    code = ISO3166Alpha2.validate('NL')
    assert code == 'NL'
    assert code.name == 'The Netherlands'


def test_name_be():
    code = ISO3166Alpha2.validate('BE')
    assert code == 'BE'
    assert code.name == 'Belgium'


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Self

import pydantic

from canonical import DomainName
from ..objectmeta import ObjectMeta
from .base import BaseNamespace
from .namespacespec import NamespaceSpec


class OrganizationSpec(NamespaceSpec):
    model_config = {'populate_by_name': True}

    id: int | None = pydantic.Field(
        default=None,
        title="Organization ID",
        description=(
            "A numeric identifier for this organization. This field is "
            "auto-generated and can not be set or changed."
        ),
        frozen=True
    )

    display_name: str = pydantic.Field(
        default=...,
        alias='displayName',
        description=(
            "The display name of the `Organization` for user "
            "interfaces."
        )
    )

    domain: DomainName | None = pydantic.Field(
        default=None,
        description=(
            "The verified domain name for this organization."
        )
    )


class Organization(BaseNamespace[ObjectMeta[str]], kind='Namespace', type='webiam.io/organization'):
    model_config = {'populate_by_name': True}

    spec: 'OrganizationSpec' = pydantic.Field(
        default=...,
        description="Defines the behavior of the `Namespace`."
    )

    def can_change(self, old: Self) -> bool:
        if not super().can_change(old):
            return False
        return all([
            self.spec.id == old.spec.id
        ])


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic

from canonical.ext.resource import Resource
from canonical.ext.resource import ObjectMetaType
from .types import Permission
from .types import PermissionWildcard

class BaseRole(
    Resource[ObjectMetaType],
    version='roles.iam.webiam.io/v1'
):
    permissions: set[PermissionWildcard | Permission] = pydantic.Field(
        default_factory=set,
        description=(
            "The set of permissions granted to this role. May contain "
            "wildcard permissions."
        )
    )

    def is_global(self) -> bool:
        return self.kind == 'ClusterRole'


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic

from canonical import Text


class UOM(pydantic.BaseModel):
    model_config = {'populate_by_name': True}    

    key: Text = pydantic.Field(
        default=...,
        description=(
            "A global identifier for the represented unit of measurement."
        )
    )

    display_name: Text = pydantic.Field(
        default=...,
        alias='displayName',
        description=(
            "The human-readable display name for the unit of "
            "measurement."
        )
    )

    @classmethod
    def each(cls):
        return cls.model_validate({
            'key': 'each',
            'display_name': 'Each'
        })


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import logging
from typing import Any
from typing import Callable
from typing import Coroutine

import fastapi
import fastapi.params
import fastapi.routing
from fastapi.dependencies.utils import get_parameterless_sub_dependant
from fastapi.exceptions import HTTPException
from starlette.routing import request_response # type: ignore

from canonical.ext.resource import Error
from canonical.utils.http import MediaTypeSelector
from .response import Response


class APIRoute(fastapi.routing.APIRoute):
    logger: logging.Logger = logging.getLogger('uvicorn')
    media_types: MediaTypeSelector = MediaTypeSelector({
        'text/html',
        'application/json',
        'application/yaml'
    })

    def inject(
        self,
        dependency: fastapi.params.Depends,
        index: int = -1
    ) -> None:
        dependant = get_parameterless_sub_dependant(
            depends=dependency,
            path=self.path_format
        )
        self.dependant.dependencies.insert(0, dependant)
        self.app = request_response(self.get_route_handler())

    def get_route_handler(self) -> Callable[[fastapi.Request], Coroutine[Any, Any, fastapi.Response]]:
        handler = super().get_route_handler()

        async def f(request: fastapi.Request) -> fastapi.Response:
            response: Response[Any]
            try:
                response = await handler(request) # type: ignore
            except HTTPException as e:
                assert request.client
                response = Response(
                    media_type=self.media_types.select(request.headers.get('Accept')),
                    status_code=e.status_code,
                    content=Error.factory({
                        'detail': e.detail,
                        'request': {
                            'url': str(request.url),
                            'host': request.client.host,
                        },
                        'status_code': e.status_code,
                    })
                )
            except Exception as e:
                self.logger.exception("Caught fatal exception.")
                assert request.client
                response = Response(
                    media_type=self.media_types.select(request.headers.get('Accept')),
                    status_code=500,
                    content=Error.factory({
                        'detail': (
                            "The server encountered an unexpected condition "
                            "that prevented it from fulfilling the request."
                        ),
                        'request': {
                            'url': str(request.url),
                            'host': request.client.host,
                        },
                        'status_code': 500,
                    })
                )
            if request.method == 'HEAD':
                # Remove the body from HEAD requests.
                response = Response(
                    status_code=response.status_code,
                    headers=response.headers,
                    media_type=response.media_type,
                )
            return response
        return f


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from __future__ import annotations
import datetime
import uuid
from typing import Any
from typing import Generic
from typing import Self
from typing import TypeVar
from typing import TYPE_CHECKING

import pydantic

from canonical.utils import deephash
if TYPE_CHECKING:
    from .resource import Resource
from .resourcekey import ResourceKey


N = TypeVar('N')
ObjectMetaType = N


class ObjectMeta(pydantic.BaseModel, Generic[N]):
    model_config = {'populate_by_name': True}
    _version: str = pydantic.PrivateAttr()
    _kind: str = pydantic.PrivateAttr()

    annotations: dict[str, Any] = pydantic.Field(
        default_factory=dict,
        title="Annotations",
        description=(
            "Annotations is an unstructured key value map stored with "
            "a resource that may be set by external tools to store "
            "and retrieve arbitrary metadata. They are not queryable and "
            "should be preserved when modifying objects."
        )
    )
    
    created: datetime.datetime | None = pydantic.Field(
        default_factory=lambda: datetime.datetime.now(datetime.timezone.utc),
        title="Created",
        description=(
            "`created` is a timestamp representing the server time "
            "when this object was created. It is not guaranteed to be set in "
            "happens-before order across separate operations. Clients may not "
            "set this value. It is represented in RFC3339 form and is in UTC. "
            "Populated by the system. Read-only. Null for lists."
        ),
        frozen=True
    )
    
    deleted: datetime.datetime | None = pydantic.Field(
        default=None,
        title="Deleted",
        description=(
            "The `deleted` field is an RFC 3339 date and time "
            "at which this resource will be deleted. This field "
            "is set by the server when a graceful deletion is "
            "requested by the user, and is not directly settable "
            "by a client. The resource is expected to be deleted "
            "(no longer visible from resource lists, and not "
            "reachable by name) after the time in this field. "
            "Once set, this value may not be unset or be set "
            "further into the future, although it may be shortened "
            "or the resource may be deleted prior to this time."
        ),
    )

    finalizers: set[str] = pydantic.Field(
        default_factory=set,
        description=(
            "Must be empty before the object is deleted from the registry. "
            "Each entry is an identifier for the responsible component that "
            "will remove the entry from the list. If the `deleted` field "
            "of the object is non-nil, entries in this list can only be "
            "removed. Finalizers may be processed and removed in any order."
        )
    )

    updated: datetime.datetime | None = pydantic.Field(
        default_factory=lambda: datetime.datetime.now(datetime.timezone.utc),
        title="Updated",
        description=(
            "`CreationTimestamp` is a timestamp representing the server time "
            "when this object was updated. Clients may not "
            "set this value. It is represented in RFC3339 form and is in UTC. "
            "Populated by the system. Read-only. Null for lists."
        )
    )

    generation: int = pydantic.Field(
        default=0,
        title="Generation",
        description=(
            "A sequence number representing a specific generation "
            "of the state. Populated by the system. Read-only."
        )
    )

    labels: dict[str, str | None] = pydantic.Field(
        default_factory=dict,
        title="Labels",
        description=(
            "Map of string keys and values that can be used to organize and "
            "categorize (scope and select) objects."
        )
    )

    name: N = pydantic.Field(
        default=...,
        title="Name",
        description=(
            "Name must be unique within a namespace. Is required when creating "
            "resources, although some resources may allow a client to request "
            "the generation of an appropriate name automatically. Name is "
            "primarily intended for creation idempotence and configuration "
            "definition. Cannot be updated."
        ),
        frozen=True
    )

    generate_name: str | None = pydantic.Field(
        default=None,
        alias='generateName',
        description=(
            "An optional prefix, used by the server, to generate a unique name "
            "ONLY IF the `name` field has not been provided. If this field "
            "is used, the name returned to the client will be different than "
            "the name passed. This value will also be combined with a unique "
            "suffix. The provided value has the same validation rules as the "
            "`name` field, and may be truncated by the length of the suffix "
            "required to make the value unique on the server. If this field "
            "is specified and the generated name exists, the server will "
            "return a 409. Applied only if `name` is not specified."
        ),
        exclude=True
    )

    resource_version: str = pydantic.Field(
        default='',
        alias='resourceVersion',
        description=(
            "An opaque value that represents the internal version of this object "
            "that can be used by clients to determine when objects have changed. "
            "May be used for optimistic concurrency, change detection, and the "
            "watch operation on a resource or set of resources. Clients must "
            "treat these values as opaque and passed unmodified back to the "
            "server. They may only be valid for a particular resource or set "
            "of resources. Populated by the system. Read-only. Value must be "
            "treated as opaque by clients and ."
        )
    )

    tags: set[str] = pydantic.Field(
        default_factory=set,
        description=(
            "An array of tags that may be used to classify an object "
            "if a label or annotation is not applicable."
        )
    )

    uid: int = pydantic.Field(
        default=0,
        title="UID",
        description=(
            "UID is the unique in time and space value for this object. "
            "It is typically generated by the server on successful creation "
            "of a resource and is not allowed to change on PUT operations. "
            "Populated by the system. Read-only."
        ),
    )

    @property
    def key(self) -> ResourceKey[N]:
        return ResourceKey(api_version=self._api_version, kind=self._kind, name=self.name)

    @classmethod
    def is_namespaced(cls) -> bool:
        return False

    @classmethod
    def default(cls) -> 'ObjectMeta[Any]':
        raise ValueError("Can not create ObjectMeta without parameters.")

    def attach(self, resource: Resource[Any]):
        self._api_version = resource.api_version
        self._kind = resource.kind

    def get_namespace(self) -> str | None:
        return None

    def is_labeled(self, names: list[str]) -> bool:
        return all([
            self.labels.get(name) not in {None, 'null'}
            for name in names
        ])

    def merge(self, metadata: Self):
        """Merge :class:`ObjectMeta` `metadata` into this instance."""
        # Only annotations, labels and tags are merged for now.
        self.annotations = {**self.annotations, **metadata.annotations}
        self.labels = {**self.labels, **metadata.labels}
        self.tags = {*self.tags, *metadata.tags}
        return self

    def update_resource_version(self, obj: pydantic.BaseModel):
        data = obj.model_dump(
            mode='json',
            exclude={'api_version', 'kind'}
        )
        self.resource_version = deephash(data, encode='hex')


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Annotated
from typing import TypeAlias

import fastapi
import httpx

from fastapi.security.utils import get_authorization_scheme_param


class RequestAuthorization(httpx.Auth):

    def __init__(self, request: fastapi.Request):
        authorization = request.headers.get("Authorization")
        self.scheme, self.credentials = get_authorization_scheme_param(authorization)

    async def async_auth_flow(self, request: httpx.Request):
        if self.scheme and self.credentials:
            request.headers['Authorization'] = f'{self.scheme} {self.credentials}'
        yield request


ImpersonationAuth: TypeAlias = Annotated[RequestAuthorization, fastapi.Depends(RequestAuthorization)]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Literal
from typing import Union

import pydantic

from canonical import DomainName
from canonical import EmailAddress
from canonical import ResourceName
from .basesubject import BaseSubject


SubjectType = Union[
    BaseSubject[Literal['Domain'],  DomainName],
    BaseSubject[Literal['Group'],  ResourceName],
    BaseSubject[Literal['ServiceAccount'], EmailAddress | ResourceName],
    BaseSubject[Literal['User'], EmailAddress | ResourceName],
]


class Subject(pydantic.RootModel[SubjectType]):
    pass


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Callable
from typing import TypeVar


T = TypeVar('T')


class ICache(str):
    __module__: str = 'canonical.protocols'

    async def get(
        self,
        key: str,
        decoder: Callable[[bytes], T] = bytes
    ) -> T | None:
        ...

    async def set(
        self,
        key: str,
        value: Any,
        encoder: Callable[..., bytes] = bytes,
        encrypt: bool = False
    ) -> None:
        """Set a key in the cache. Encode `value` using the given
        encode if it is not a byte-sequence.
        """
        ...


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import inspect
from collections import OrderedDict
from typing import Any
from typing import Callable

from ._deephash import deephash
from .deferred import Deferred
from .loader import import_symbol


__all__: list[str] = [
    'deephash',
    'import_symbol',
    'Deferred',
]


class class_property:
    __module__: str = 'oauthx.utils'

    def __init__(self, func: Callable[..., Any]):
        self.func = func

    def __get__(self, instance: Any, cls: Any) -> Any:
        return self.func(cls)


def merge_signatures(signatures: list[inspect.Signature]) -> inspect.Signature:
    """Merge signatures to that FastAPI can inject the dependencies."""
    params: dict[str, inspect.Parameter] = OrderedDict()
    for sig in signatures:
        for param in sig.parameters.values():
            if param.name in {'self', 'cls'}:
                continue
            if param.name.startswith('_'):
                continue
            if param.kind == inspect.Parameter.VAR_POSITIONAL:
                continue
            if param.kind == inspect.Parameter.VAR_KEYWORD:
                continue
            params[param.name] = param

    return signatures[0].replace(
        parameters=list(sorted(params.values(), key=lambda p: (p.kind, p.default != inspect._empty))) # type: ignore
    )


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import urllib.parse
from typing import Any

from pydantic import GetJsonSchemaHandler
from pydantic.json_schema import JsonSchemaValue
from pydantic_core import CoreSchema
from pydantic_core import core_schema

from .stringrepresentable import StringRepresentable


class HTTPResourceLocator(StringRepresentable):

    @classmethod
    def __get_pydantic_core_schema__(cls, *_: Any) -> CoreSchema:
        return core_schema.json_or_python_schema(
            json_schema=core_schema.str_schema(max_length=2048),
            python_schema=core_schema.union_schema([
                core_schema.is_instance_schema(cls),
                core_schema.chain_schema([
                    core_schema.no_info_plain_validator_function(str),
                    core_schema.str_schema(max_length=2048),
                    core_schema.no_info_plain_validator_function(cls.fromstring)
                ])
            ]),
            serialization=core_schema.plain_serializer_function_ser_schema(str)
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls,
        _: CoreSchema,
        handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue: # pragma: no cover
        return handler(core_schema.str_schema(max_length=2048))

    @classmethod
    def fromstring(cls, v: str, _: Any = None):
        p = urllib.parse.urlparse(v)
        if p.scheme not in {'http', 'https'}:
            raise ValueError(f"Not a valid URL: {v[:128]}")
        return cls(v)

    def __init__(self, value: str):
        self.__value = value
        
    def __str__(self) -> str:  # pragma: no cover
        return self.__value


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .acceptedcontenttype import AcceptedContentType
from .authorizationservice import AuthorizationService
from .localauthorization import LocalAuthorizationContext
from .negotiatedresponsemediatype import NegotiateResponseMediaType
from .remoteauthorization import RemoteAuthorizationContext
from .requestauthorizationcontext import RequestAuthorizationContext
from .requestemail import RequestEmail
from .requestverb import RequestVerb
from .resourcerepository import ResourceRepository
from .requestresource import RequestResource
from .requestresourcename import RequestResourceName
from .resourcemodel import ResourceModel


__all__: list[str] = [
    'AcceptedContentType',
    'AuthorizationService',
    'LocalAuthorizationContext',
    'NegotiateResponseMediaType',
    'RemoteAuthorizationContext',
    'RequestAuthorizationContext',
    'RequestEmail',
    'RequestResource',
    'RequestResourceName',
    'RequestVerb',
    'ResourceModel',
    'ResourceRepository'
]


# Copyright (C) 2023-2024 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Generic
from typing import TypeVar

import pydantic


__all__: list[str] = [
    'ResourceID'
]

N = TypeVar('N')
T = TypeVar('T')


class ResourceID(pydantic.BaseModel, Generic[N, T]):
    """A container to reference an `id` for any resource type. A resource "
    is a generic term for something you (a developer) may want to interact "
    "with through one of our API's.
    """
    type: N = pydantic.Field(
        default=...,
        description="The resource type this `ResourceID` is for."
    )

    name: T = pydantic.Field(
        default=...,
        description=(
            "The type-specific id. This should correspond to "
            "the id used in the type-specific API's."
        )
    )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic

from .emailaddress import EmailAddress
from .honorific import HonorificEnum
from .phonenumber import Phonenumber


class DeliveryContact(pydantic.BaseModel):
    """Describes a person that may be contacted in the context of a
    delivery.
    """
    model_config = {'populate_by_name': True}

    name: str = pydantic.Field(
        default=...
    )

    email: EmailAddress | None = pydantic.Field(
        default=None
    )

    phonenumber: Phonenumber | str | None = pydantic.Field(
        default=None
    )
    
    honorific: HonorificEnum | None = pydantic.Field(
        default=None
    )

    @property
    def display_name(self) -> str:
        return self.name


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

import pydantic

from canonical.utils.http import Request
from .resource import Resource
from .transientmeta import TransientMeta


class ErrorData(pydantic.BaseModel):
    model_config = {'populate_by_name': True}

    detail: str = pydantic.Field(
        default=...,
        description="A summary of the error."
    )


class HTTPErrorData(ErrorData):
    status_code: int = pydantic.Field(
        default=...,
        alias='statusCode',
        description=(
            "The HTTP status code identifying the error state."
        )
    )

    request: Request = pydantic.Field(
        default=...,
        description=(
            "Describes the HTTP request that caused the error "
            "condition."
        )
    )


class Error(Resource[TransientMeta], version='errors/v1'):
    data: HTTPErrorData

    @classmethod
    def factory(cls, data: dict[str, Any]):
        return cls.model_validate({
            'data': data
        })


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import ClassVar
from typing import TypeVar

from canonical.ext.resource import ResourceType
from .base import BaseOperation


T = TypeVar('T')


class Default(BaseOperation[T]):
    verb: ClassVar[str]

    def __init__(self, model: type[ResourceType], **kwargs: Any):
        super().__init__(verb=self.verb, model=model, **kwargs)


# Copyright (C) 2023-2024 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Generic
from typing import TypeVar

import pydantic


__all__: list[str] = [
    'ObjectReference'
]

G = TypeVar('G')
K = TypeVar('K')
T = TypeVar('T')


class ObjectReference(pydantic.BaseModel, Generic[G, K, T]):
    model_config = {'populate_by_name': True}

    api_group: G = pydantic.Field(
        default=None,
        alias='apiGroup',
        title="API group",
        description=(
            "Specifies the API group of the referent. Cannot be updated."
        ),
        frozen=True
    )

    kind: K = pydantic.Field(
        default=...,
        description=(
            "Kind of the referent. Cannot be updated. In `CamelCase`."
        ),
        frozen=True
    )

    name: T = pydantic.Field(
        default=...,
        description=(
            "The `.metadata.name` of of the referent. Cannot be "
            "updated."
        ),
        frozen=True
    )


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import functools
import logging
from typing import cast
from typing import Any
from typing import AsyncIterable

from google.cloud.datastore import Client

from canonical.protocols import ITransaction
from canonical.protocols import ITyped
from canonical.ext.resource import IResourceRepository
from canonical.ext.resource import PrimaryKey
from canonical.ext.resource import ResourceType
from canonical.ext.resource import ResourceTypeVar
from .basedatastorestorage import BaseDatastoreStorage
from .protocols import IDatastoreKey
from .protocols import IDatastoreEntity


class ResourceRepository(BaseDatastoreStorage, IResourceRepository):
    cluster_namespace: str
    logger: logging.Logger = logging.getLogger('uvicorn')

    async def allocate(self, obj: type[ResourceType]) -> int:
        k = f'{obj.group}/{obj.__name__}'
        i = await self.allocate_identifier(k)
        self.logger.info("Allocated identifier (kind: %s, uid: %s)", k, i)
        return i

    def all(
        self,
        model: type[ResourceTypeVar],
        namespace: str | None = None
    ) -> AsyncIterable[ResourceTypeVar]:
        return self.query( # type: ignore
            model=model,
            namespace=namespace
        )

    def get_entity_name(self, cls: type[ResourceType]) -> str:
        name = f'{cls.group}/{cls.__name__}'
        if not cls.group:
            name = cls.__name__
        return name

    def resource_key(self, resource: ResourceType, model: type[ResourceType] | None = None) -> IDatastoreKey:
        args: list[Any] = [self.get_entity_name(model or type(resource)), resource.metadata.name]
        return self.client.key(*args, namespace=resource.get_namespace()) # type: ignore

    async def exists(self, key: ITyped[ResourceType]) -> bool:
        assert isinstance(key, PrimaryKey)
        filters: list[tuple[str, str, Any]] = [
            ('metadata.name', '=', key.name)
        ]
        if key.namespace is not None:
            filters.append(('metadata.namespace', '=', key.namespace))
        q = self.query(model=key.model, filters=filters)
        return await q.exists()

    async def get(
        self,
        key: ITyped[Any],
        cached: bool = False,
        model: type[ResourceType] | None = None,
        max_age: int = 0,
        transaction: ITransaction | None = None,
        **kwargs: Any
    ) -> ResourceType:
        key = cast(PrimaryKey[ResourceType], key)
        if cached:
            raise NotImplementedError
        obj = await self.get_model_by_key(
            cls=key.get_type(),
            pk=key.name
        )
        if obj is None:
            raise self.DoesNotExist
        return obj

    async def get_by_name(
        self,
        model: type[ResourceType],
        name: str | int,
        namespace: str | None = None
    ) -> ResourceType:
        obj = await self.get_model_by_key(model, name, namespace=namespace)
        if obj is None:
            raise self.DoesNotExist
        return obj

    async def persist_entity(
        self,
        client: Client | ITransaction,
        entity: IDatastoreEntity
    ) -> IDatastoreEntity:
        return await self.run_in_executor(functools.partial(client.put, entity)) # type: ignore

    async def persist(
        self,
        object: ResourceType,
        transaction: ITransaction | None = None,
        model: type[ResourceType] | None = None,
        **kwargs: Any
    ) -> ResourceType:
        entity = self.entity_factory(self.resource_key(object, model=model), object)
        await self.persist_entity(transaction or self.client, entity)
        return object


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Protocol

from canonical.protocols import ITransaction
from .idatastoreentity import IDatastoreEntity


class IDatastoreTransaction(ITransaction, Protocol):
    __module__: str = 'cbra.ext.google.types'

    def begin(self) -> None: ...
    def commit(self) -> None: ...
    def put(self, entity: IDatastoreEntity) -> None: ...
    def rollback(self) -> None: ...


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Generic
from typing import TypeVar

import pydantic

from ..resourceid import ResourceID


N = TypeVar('N')
T = TypeVar('T')


class NamespaceSpec(pydantic.BaseModel):
    finalizers: set[str] = pydantic.Field(
        default_factory=list,
        description=(
            "Finalizers is an opaque list of values that must "
            "be empty to permanently remove object from storage."
        )
    )


class HierarchicalNamespaceSpec(NamespaceSpec, Generic[N, T]):
    parent: ResourceID[N, T] | None = pydantic.Field(
        default=None,
        description=(
            "An optional reference to a parent Resource.\n\nSupported parent "
            "types include `organization` and `folder`. Once set, the parent "
            "cannot be cleared. The parent can be set on creation or "
            "using the `projects.update` method; the end user must "
            "have the `projects.create` permission on the parent."
        )
    )


# Copyright (C) 2023-2024 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import ClassVar
from typing import Generic
from typing import TypeVar
from typing import TYPE_CHECKING

from canonical.protocols import ITyped
if TYPE_CHECKING:
    from .rootresource import ResourceType


__all__: list[str] = [
    'PrimaryKey'
]

M = TypeVar('M')


class PrimaryKey(ITyped[M], Generic[M]):
    model: type[M]
    api_group: ClassVar[str]
    plural: ClassVar[str]
    name: str
    namespace: str | None

    @classmethod
    def typed(cls, t: type['ResourceType']) -> type[ITyped[M]]:
        return type(f'{t.__name__}PrimaryKey', (cls,), {
            'api_group': t.group,
            'model': t,
            'plural': t.plural,
        })

    def __init__(
        self,
        name: str,
        namespace: str | None = None,
        server: str | None = None
    ):
        self.name = name
        self.namespace = namespace
        self.server = server

    def get_type(self) -> type[M]:
        return self.model

    def __repr__(self):
        return f"<Key('{self.model.__name__}', {repr(self.name)}, namespace={repr(self.namespace)})"


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import inspect
from types import ModuleType
from typing import get_args
from typing import get_origin
from typing import Any
from typing import Iterable
from typing import Union

import fastapi

from canonical.ext.resource import APIResourceList
from canonical.ext.resource import Error
from canonical.ext.resource import Resource
from canonical.ext.resource import ResourceInspector
from canonical.ext.resource import ResourceMeta
from canonical.ext.resource import ResourceType
from canonical.ext.resource import RootResource
from canonical.utils.http import MediaTypeSelector

from .params import NegotiateResponseMediaType
from .resourcerouter import ResourceRouter
from .response import Response


class ResourceApplication(fastapi.FastAPI):
    inspector: ResourceInspector = ResourceInspector()
    resource_paths: set[str] = set()
    media_types: MediaTypeSelector = MediaTypeSelector({
        'text/html',
        'application/json',
        'application/yaml'
    })

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.apis: dict[str, dict[str, APIResourceList]] = {}

        @self.exception_handler(404)
        def _(request: fastapi.Request, _: Any) -> Any:
            assert request.client is not None
            return Response(
                media_type=self.media_types.select(request.headers.get('Accept')),
                status_code=404,
                content=Error.factory({
                    'status_code': 404,
                    'detail': "The server cannot find the requested resource",
                    'request': {
                        'url': str(request.url),
                        'host': request.client.host
                    }
                })
            )

    def add(
        self,
        impl: type[Resource[Any] | RootResource[Any] | ResourceRouter[Any]] | ModuleType,
        verbs: Iterable[str] | None = None,
        **kwargs: Any
    ):
        self.openapi_tags = self.openapi_tags or []
        verbs = set(verbs or [])
        if inspect.isclass(impl) and issubclass(impl, ResourceRouter):
            router: ResourceRouter[Any] = impl(enabled_verbs=verbs)
            router.add_to_application(self)
            self.include_router(router=router)
        elif get_origin(impl) == Union:
            for model in get_args(impl):
                self.add(model, **kwargs)
        elif isinstance(impl, ModuleType):
            for _, value in inspect.getmembers(impl):
                if not isinstance(value, ResourceRouter):
                    continue
                value.add_to_application(self)
        else:
            router = ResourceRouter.with_model(self, impl, **kwargs) # type: ignore
            self.include_router(router=router) # type: ignore

    def create_discovery_endpoint(
        self,
        meta: ResourceMeta,
        resources: APIResourceList,
    ) -> None:

        async def discover(
            media_type: str = NegotiateResponseMediaType({'application/yaml', 'application/json', "text/html"})
        ) -> Response[Any]:
            return Response(
                status_code=200,
                media_type=media_type,
                content=self.apis[meta.api_group][meta.version]
            )

        self.add_api_route(
            methods=['GET'],
            path=f'/{resources.path(meta.api_group)}',
            endpoint=discover,
            summary="List the resources that this server supports",
            response_description="The list of resources that the server supports.",
            response_model=APIResourceList,
            include_in_schema=True,
            tags=['Service Endpoints']
        )
        

    def register_resource(self, model: type[ResourceType], verb: str) -> None:
        meta = self.inspector.inspect(model)
        if meta.api_group not in self.apis:
            self.apis[meta.api_group] = {}
        if meta.version not in self.apis[meta.api_group]:
            resources = self.apis[meta.api_group][meta.version] = APIResourceList(
                groupVersion=meta.version
            )
            self.create_discovery_endpoint(meta, resources)
        resource = model.register(self.apis[meta.api_group][meta.version])
        resource.add(verb)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

from ..rootresource import RootResource
from .domain import Domain
from .organization import Organization
from .project import Project
from .namespace import Namespace as NamespaceType


class Namespace(RootResource[Organization | Project | Domain | NamespaceType]):
    """A `Namespace` provides a mechanism for isolating groups of resources.
    Names of resources need to be unique within a namespace, but not across
    namespaces. Namespace-based scoping is applicable only for namespaced
    objects and not for system-wide objects
    """
    model_config = {'populate_by_name': True}

    @classmethod
    def factory(cls, type: str, name: str, **extra: Any):
        params: dict[str, Any] = {
            'metadata': {
                'name': name
            },
            'type': type
        }
        return cls.model_validate({**extra, **params})

    @classmethod
    def is_namespaced(cls) -> bool:
        return False

    @property
    def spec(self):
        return self.root.spec

    @property
    def type(self):
        return self.root.type

    def is_organization(self) -> bool:
        return self.root.type == 'webiam.io/organization'


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Iterable

import fastapi
import fastapi.params
from fastapi.exceptions import HTTPException

from canonical.utils.http import MediaTypeSelector


def NegotiateResponseMediaType(available: Iterable[str]) -> Any:
    selector = MediaTypeSelector(available)

    def f(request: fastapi.Request):
        header = request.headers.get('Accept')
        if header is None:
            raise HTTPException(
                status_code=406,
                detail="The Accept header is required."
            )
        selected = selector.select(header)
        if selected is None:
            raise HTTPException(
                status_code=406,
                detail=(
                    "Unable to select a response content type based on "
                    "the provided Accept header."
                )
            )
        setattr(request.state, 'media_type', selected)
        return selected

    return fastapi.Depends(f)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Literal

import pydantic

from canonical import DomainName
from ..resourceid import ResourceID
from ..objectmeta import ObjectMeta
from .base import BaseNamespace
from .namespacespec import NamespaceSpec


class Domain(BaseNamespace[ObjectMeta[DomainName]], kind='Namespace', type='webiam.io/domain'):
    model_config = {'populate_by_name': True}

    spec: 'DomainSpec' = pydantic.Field(
        default_factory=NamespaceSpec,
        description="Defines the behavior of the `Namespace`."
    )


class DomainSpec(NamespaceSpec):
    parent: ResourceID[Literal['organization'], int] = pydantic.Field(
        default=...,
        description=(
            "Defines the organization the domain is verified for."
        )
    )


Domain.model_rebuild()


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Generic
from typing import TypeVar

import pydantic

from .objectmeta import ObjectMeta
from .objectmeta import N


T = TypeVar('T')


class NamespacedObjectMeta(ObjectMeta[N], Generic[N]):
    namespace: str = pydantic.Field(
        default=...,
        title="Namespace",
        description=(
            "Namespace defines the space within which each name must "
            "be unique. An empty namespace is equivalent to the `default` "
            "namespace, but `default` is the canonical representation. "
            "Not all objects are required to be scoped to a namespace - "
            "the value of this field for those objects will be empty. "
            "Must be a DNS_LABEL. Cannot be updated."
        )
    )

    @classmethod
    def is_namespaced(cls) -> bool:
        return True

    def get_namespace(self) -> str | None:
        return self.namespace


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from __future__ import annotations
import datetime
import uuid
from typing import Any
from typing import TYPE_CHECKING

import pydantic

if TYPE_CHECKING:
    from .resource import Resource
from .resourcekey import ResourceKey


class TransientMeta(pydantic.BaseModel):
    model_config = {'populate_by_name': True}

    annotations: dict[str, Any] = pydantic.Field(
        default_factory=dict,
        title="Annotations",
        description=(
            "Annotations is an unstructured key value map stored with "
            "a resource that may be set by external tools to store "
            "and retrieve arbitrary metadata. They are not queryable and "
            "should be preserved when modifying objects."
        )
    )
    
    created: datetime.datetime | None = pydantic.Field(
        default_factory=lambda: datetime.datetime.now(datetime.timezone.utc),
        title="Created",
        description=(
            "`created` is a timestamp representing the server time "
            "when this object was created. It is not guaranteed to be set in "
            "happens-before order across separate operations. Clients may not "
            "set this value. It is represented in RFC3339 form and is in UTC. "
            "Populated by the system. Read-only. Null for lists."
        ),
        frozen=True
    )

    labels: dict[str, str | None] = pydantic.Field(
        default_factory=dict,
        title="Labels",
        description=(
            "Map of string keys and values that can be used to organize and "
            "categorize (scope and select) objects."
        )
    )

    tags: set[str] = pydantic.Field(
        default_factory=set,
        description=(
            "An array of tags that may be used to classify an object "
            "if a label or annotation is not applicable."
        )
    )

    uid: str = pydantic.Field(
        default_factory=lambda: str(uuid.uuid4()),
        title="UID",
        description=(
            "UID is the unique in time and space value for this object. "
            "It is typically generated by the server on successful creation "
            "of a resource and is not allowed to change on PUT operations. "
            "Populated by the system. Read-only."
        ),
        frozen=True
    )

    @classmethod
    def is_namespaced(cls) -> bool:
        return False

    @classmethod
    def default(cls):
        return cls()

    @property
    def name(self) -> str:
        raise NotImplementedError

    @property
    def key(self) -> ResourceKey[Any]:
        raise NotImplementedError

    @property
    def resource_version(self) -> None:
        return None

    def attach(self, resource: Resource[Any]):
        self._api_version = resource.api_version
        self._kind = resource.kind

    def get_namespace(self) -> str | None:
        return None

    def update_resource_version(self, obj: Resource[Any]):
        pass


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import fastapi

from canonical.ext.resource import IResourceRepository
from .resourceapplication import ResourceApplication
from .resourceclient import ResourceClient
from .resourcerouter import ResourceRouter


__all__: list[str] = [
    'ResourceApplication',
    'ResourceClient',
    'ResourceRouter'
]

def setup_dependencies(
    repository: type[IResourceRepository]
):
    async def f(
        request: fastapi.Request,
        resources: IResourceRepository = fastapi.Depends(repository)
    ):
        setattr(request.state, 'resources', resources)

    return [fastapi.Depends(f)]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .canonicalexception import CanonicalException
from .duplicate import Duplicate
from .doesnotexist import DoesNotExist
from .immutable import Immutable
from .multipleobjectsreturned import MultipleObjectsReturned
from .programmingerror import ProgrammingError
from .stale import Stale


__all__: list[str] = [
    'CanonicalException',
    'Duplicate',
    'DoesNotExist',
    'Immutable',
    'MultipleObjectsReturned',
    'ProgrammingError',
    'Stale',
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import asyncio
import contextlib
import functools
import inspect
from typing import cast
from typing import Any
from typing import AsyncIterable
from typing import Callable
from typing import Iterable
from typing import Mapping
from typing import TypeVar

import pydantic
from google.cloud.datastore import Client

from canonical import PersistedModel
from canonical import ResourceIdentifier
from canonical.utils import class_property
from canonical.utils import merge_signatures
from .datastorecursor import DatastoreCursor
from .protocols import IDatastoreKey
from .protocols import IDatastoreEntity
from .protocols import IDatastoreTransaction


T = TypeVar('T', bound=pydantic.BaseModel)


class BaseDatastoreStorage:
    dump_mode: str = 'json'
    exclude_fields: set[str] = set()
    exclude_none: bool = False

    @class_property
    def __signature__(cls) -> inspect.Signature:
        return merge_signatures([
            inspect.signature(cls.__init__),
            inspect.signature(cls.require)
        ])

    @staticmethod
    async def run_in_executor(
        func: Callable[..., Any],
        *args: Any,
        **kwargs: Any
    ) -> Any:
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(None, func, *args, **kwargs)

    def __init__( # type: ignore
        self,
        *,
        client: Client,
        **kwargs: Any
    ) -> None:
        self.client = client
        self.require(**kwargs)

    def require(self, **_: Any):
        pass

    async def allocate_identifier(self, cls: type | str) -> int:
        if not isinstance(cls, str):
            cls = cls.__name__
        base = self.entity_key(kind=cls)
        result = await self.run_in_executor(
            functools.partial( # type: ignore
                self.client.allocate_ids, # type: ignore
                incomplete_key=base,
                num_ids=1
            )
        )
        return [x for x in result][0].id

    async def delete(self, obj: Any) -> None:
        key = self.entity_key(type(obj), obj.pk)
        await self.run_in_executor(
            functools.partial(
                self.client.delete, # type: ignore
                key=key
            )
        )

    def dump_model(
        self,
        obj: pydantic.BaseModel,
        exclude_fields: set[str] | dict[str, Any] | None = None,
        exclude_none: bool = False
    ) -> dict[str, Any]:
        return obj.model_dump(
            mode=self.dump_mode,
            exclude=exclude_fields or self.exclude_fields,
            exclude_none=exclude_none or self.exclude_none
        )

    def entity_factory(
        self,
        key: IDatastoreKey,
        obj: pydantic.BaseModel,
        exclude_fields: set[str] | dict[str, Any] | None = None
    ) -> IDatastoreEntity:
        entity = self.client.entity(key) # type: ignore
        attrs = self.dump_model(obj, exclude_fields=exclude_fields)
        if attrs:
            entity.update(attrs) # type: ignore
        return entity # type: ignore

    def entity_key(
        self,
        kind: str | type,
        identifier: int | str | None = None,
        parent: IDatastoreKey | None = None,
        namespace: str | None = None
    ) -> IDatastoreKey:
        if not isinstance(kind, str):
            kind = self.get_entity_name(kind)
        if isinstance(identifier, ResourceIdentifier):
            identifier = identifier.cast() # type: ignore
        args: list[Any] = [kind]
        if identifier is not None:
            args.append(identifier)
        return self.client.key(*args, parent=parent, namespace=namespace) # type: ignore

    def get_entity_name(self, cls: type[Any]) -> str:
        return cls.__name__

    def model_factory(self, entity: Mapping[str, Any], model: type[T]) -> T:
        obj = model.model_validate(dict(entity))
        if isinstance(obj, PersistedModel):
            obj.attach(self) # type: ignore
        return obj

    def query(
        self,
        model: type[T],
        filters: Iterable[tuple[str, str, Any]] | None = None,
        sort: Iterable[str] | None = None,
        namespace: str | None = None,
        limit: int | None = None,
        kind: str | None = None,
        page_size: int = 10,
        keys: list[IDatastoreKey] | None = None,
        **_: Any
    ) -> DatastoreCursor[T]:
        flattened: list[tuple[str, str, Any]] = []
        for attname, op, value in list(filters or []):
            if not isinstance(value, list):
                flattened.append((attname, op, value))
                continue
            flattened.extend([(attname, op, x) for x in value]) # type: ignore
        return DatastoreCursor(
            kind=kind or self.get_entity_name(model),
            model=model,
            client=self.client,
            keys=keys,
            filters=flattened,
            namespace=namespace,
            sort=sort,
            limit=limit,
            page_size=page_size
        )

    async def find(
        self,
        kind: type[T],
        filters: Iterable[tuple[str, str, int | str]],
        sort: Iterable[str] | None = None
    ) -> AsyncIterable[T]:
        q = self.client.query(kind=kind.__name__) # type: ignore
        if sort is not None:
            q.order = sort
        for filter in filters:
            q.add_filter(*filter) # type: ignore
        results = await self.run_in_executor(
            functools.partial(q.fetch) # type: ignore
        )
        for entity in results:
            yield kind.model_validate(dict(entity))

    async def get_entity_by_key(self, key: IDatastoreKey) -> IDatastoreEntity | None:
        return await self.run_in_executor(
            functools.partial(
                self.client.get, # type: ignore
                key=key
            )
        )

    async def get_model_by_key(
        self,
        cls: type[T],
        pk: int | str,
        parent: IDatastoreKey | None = None,
        namespace: str | None = None
    ) -> T | None:
        entity = await self.get_entity_by_key(
            key=self.entity_key(cls, pk, parent=parent, namespace=namespace)
        )
        if entity is None:
            return None
        return await self.restore(self.model_factory(dict(entity), cls))

    async def put(
        self,
        obj: pydantic.BaseModel,
        pk: int | str,
        parent: IDatastoreKey | None = None,
        exclude_fields: set[str] | dict[str, Any] | None = None,
        transaction: IDatastoreTransaction | None = None
    ) -> IDatastoreKey:
        put = self.client.put if transaction is None else transaction.put # type: ignore
        key = self.entity_key(type(obj), pk, parent=parent)
        entity = self.entity_factory(key, obj, exclude_fields=exclude_fields)
        await self.run_in_executor(put, entity) # type: ignore
        return entity.key # type: ignore

    async def restore(self, instance: T) -> T:
        return instance 

    @contextlib.asynccontextmanager
    async def transaction(self, transaction: IDatastoreTransaction | None = None):
        tx = cast(IDatastoreTransaction, await self.run_in_executor(self.client.transaction)) # type: ignore
        try:
            await self.run_in_executor(tx.begin)
            yield tx
            await self.run_in_executor(tx.commit)
        except Exception:
            await self.run_in_executor(tx.rollback)
            raise


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import copy
import inspect
import logging
from collections import OrderedDict
from typing import cast
from typing import Any
from typing import Callable
from typing import Generic
from typing import NoReturn
from typing import TypeVar

import fastapi
import pydantic
from fastapi.exceptions import HTTPException

from canonical.exceptions import Duplicate
from canonical.exceptions import Immutable
from canonical.ext.resource import Error
from canonical.ext.resource import ResourceType
from canonical.ext.resource import IResourceRepository
from canonical.ext.iam.protocols import IResourceAuthorizationContext
from ..params import AcceptedContentType
from ..params import RequestAuthorizationContext
from ..params import RequestResourceName
from ..params import RequestVerb
from ..params import ResourceRepository
from .endpoint import Endpoint


T = TypeVar('T')


class BaseOperation(Generic[T]):
    creates: bool = False
    dependants: list[Callable[..., Any]]
    dependencies: list[Any]
    input_model: type[pydantic.BaseModel] | None = None
    logger: logging.Logger = logging.getLogger('uvicorn')
    method: str
    model: type[ResourceType]
    permissions: set[str]

    default_responses: dict[int, Any] = {
        401: {
            'model': Error,
            'description': (
                "Authentication is required to perform the requested "
                "operation or the provided credential is invalid."
            )
        },
        403: {
            'model': Error,
            'description': (
                "Untrusted credential or the authenticated request is not allowed "
                "to perform the requested operation."
            )
        },
        406: {
            'model': Error,
            'description': (
                "The media type accepted by the client can not be "
                "satisfied by the server."
            )
        },
        500: {
            'model': Error,
            'description': (
                "The server encountered an unexpected condition that "
                "prevented it from fulfilling the request."
            )
        }
    }
    detail: bool = False
    existing: bool = False
    requires_body: bool = False
    path: str | None = None
    status_code: int = 200

    def __init__(
        self,
        verb: str,
        model: type[ResourceType],
        permissions: set[str] | None = None
    ) -> None:
        self.model = model
        self.verb = verb
        self.dependants = []
        self.dependencies = [
            fastapi.Depends(self.inject_verb(verb)),
            fastapi.Depends(self.inject_model),
        ]
        self.permissions = set(permissions or [])
        if self.is_namespaced():
            self.dependencies.append(fastapi.Depends(self.inject_namespace))
        if self.is_detail():
            self.dependencies.append(fastapi.Depends(self.inject_name))
        if self.needs_existing():
            self.dependencies.append(fastapi.Depends(self.inject_existing))
        if self.input_model is not None:
            self.requires_body = True
        if self.requires_body:
            self.dependencies.append(
                AcceptedContentType({'application/yaml', 'application/json'}),
            )

    def add_to_router(
        self,
        router: fastapi.FastAPI | fastapi.APIRouter,
        register: Callable[[type[ResourceType], str], None]
    ):
        router.add_api_route(
            methods=[self.method],
            path=self.get_path(),
            endpoint=self.as_handler(),
            name=self.get_endpoint_name(),
            summary=self.get_endpoint_summary(),
            description=getattr(self.handle, '__doc__', None),
            dependencies=[*self.dependencies, fastapi.Depends(self.authorize)],
            responses={**self.get_openapi_responses()},
            response_description=self.get_response_description(),
            response_model=self.get_response_model(),
            response_model_by_alias=True,
            status_code=self.status_code,
            tags=[self.model.__name__],
            operation_id=self.get_endpoint_name()
        )
        register(self.model, self.verb)

    def as_handler(self):
        # Do some modifications on the signatures because for
        # the base implementation classes, not every type is
        # known yet (i.e. the actual model).
        sig = inspect.signature(self.handle)
        params: OrderedDict[str, inspect.Parameter] = OrderedDict(sig.parameters)
        for param in sig.parameters.values():
            if param.annotation != ResourceType:
                continue
            params[param.name] = param.replace(annotation=self.get_model())

        if not params.get('request'):
            params['request'] = inspect.Parameter(
                kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
                name='request',
                annotation=fastapi.Request
            )

        return Endpoint(
            name=self.get_endpoint_name(),
            signature=sig.replace(parameters=list(params.values())),
            handle=self,
        )

    def depends(self, func: Callable[..., Any]) -> Callable[..., Any]:
        self.dependants.append(func)
        return func

    def fail(self, status_code: int, detail: str) -> NoReturn:
        raise HTTPException(status_code=status_code, detail=detail)

    def get_endpoint_name(self) -> str:
        return f'{self.model.plural}.{self.verb}'

    def get_endpoint_summary(self) -> str:
        raise NotImplementedError

    def get_model(self) -> type[pydantic.BaseModel]:
        return self.model

    def get_openapi_responses(self) -> dict[int, dict[str, Any]]:
        responses = copy.deepcopy(self.default_responses)
        if self.detail:
            responses[404] = {
                'model': Error,
                'description': (
                    f'The {self.model.__name__} specified by the path parameter(s) '
                    'does not exist.'
                )
            }
        if self.requires_body:
            responses[415] = {
                'model': Error,
                'description': "Invalid content type for request body."
            }
        if self.method == 'HEAD':
            responses.pop(406)
        return responses

    def get_path(self) -> str:
        path = f'{self.model.base_path}'
        if self.detail:
            path = f'{path}/{{name}}'
        return f'/{path}'

    def get_response_description(self) -> str:
        raise NotImplementedError

    def get_response_model(self) -> type[pydantic.BaseModel] | None:
        if self.method == 'HEAD':
            return None
        return self.model if (self.detail or self.creates) else self.model.List

    def inject_name(
        self,
        request: fastapi.Request,
        name: str = fastapi.Path(
            description=f'The `.metadata.name` of an existing resource.',
            max_length=64,
        )
    ):
        setattr(request.state, 'name', name)

    def inject_namespace(
        self,
        request: fastapi.Request,
        namespace: str = fastapi.Path(
            description=(
                "Identifies the namespace that contains the "
                "resource."
            )
        ),
    ):
        setattr(request.state, 'namespace', namespace)

    def inject_model(self, request: fastapi.Request):
        setattr(request.state, 'model', self.model)

    def inject_verb(self, verb: str):
        def f(request: fastapi.Request):
            setattr(request.state, 'verb', verb)
        return f

    def is_detail(self) -> bool:
        return all([
            not self.creates,
            not self.detail
        ])

    def is_namespaced(self) -> bool:
        return self.model.is_namespaced()

    def needs_existing(self) -> bool:
        return self.existing

    def on_not_authenticated(self):
        raise HTTPException(
            status_code=401,
            detail="Authentication required."
        )

    def on_unauthorized(self):
        raise HTTPException(
            status_code=403,
            detail=(
                "The request subject is not granted permission "
                "to perform this operation."
            )
        )

    async def authorize(
        self,
        ctx: RequestAuthorizationContext,
        verb: RequestVerb
    ) -> None:
        if verb == 'authorize':
            return
        if not ctx.is_authenticated():
            self.on_not_authenticated()
        if not ctx.is_authorized():
            self.on_unauthorized()

        if self.permissions and not await ctx.has(self.permissions):
            raise NotImplementedError

    async def handle(self, *args: Any, **kwargs: Any) -> T:
        raise NotImplementedError

    async def inject_existing(
        self,
        request: fastapi.Request,
        ctx: RequestAuthorizationContext,
        resources: ResourceRepository,
        key: RequestResourceName
    ) -> None:
        try:
            obj = await resources.get(key)
            assert obj is not None
        except resources.DoesNotExist:
            self.fail(404, f"{self.model.__name__} does not exist.")
        await self.on_mutation_request(request, ctx, resources, cast(T, obj))
        setattr(request.state, 'resource', obj)

    async def on_mutation_request(
        self,
        request: fastapi.Request,
        ctx: IResourceAuthorizationContext,
        repo: IResourceRepository,
        resource: T
    ) -> None:
        """Hook to implement logic prior to mutating a resource."""
        pass

    async def render_to_response(self, result: T) -> fastapi.Response:
        raise NotImplementedError

    async def __call__(
        self,
        request: fastapi.Request,
        *args: Any, **kwargs: Any
    ) -> fastapi.Response:
        sig = inspect.signature(self.handle)
        if sig.parameters.get('request'):
            kwargs['request'] = request
        try:
            result = await self.handle(**{
                k: v for k ,v in kwargs.items()
                if k in sig.parameters
            })
        except Duplicate:
            self.fail(409, "Resource conflict.")
        except Immutable as e:
            self.fail(409, e.detail)
        return await self.render_to_response(result)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .baseauthorizationcontext import BaseAuthorizationContext
from .clusterrole import ClusterRole
from .clusterrolebinding import ClusterRoleBinding
from .permissionquery import PermissionQuery
from .role import Role
from .rolebinding import RoleBinding
from .permissionset import PermissionSet


__all__: list[str] = [
    'BaseAuthorizationContext',
    'ClusterRole',
    'ClusterRoleBinding',
    'PermissionQuery',
    'PermissionSet',
    'Role',
    'RoleBinding',
]


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import re
from typing import Any
from typing import Callable
from typing import Generator

from pydantic_core import CoreSchema
from pydantic_core import core_schema
from pydantic.json_schema import JsonSchemaValue
from pydantic import GetJsonSchemaHandler


class StringType(str):
    """Base class for string types."""
    __module__: str = 'canonical'
    openapi_title: str | None = None
    openapi_format: str | None = None
    max_length: int | None = None
    min_length: int | None = None
    patterns: re.Pattern[Any] | list[re.Pattern[Any]] = []
    __value: str

    @classmethod
    def __get_pydantic_json_schema__(
        cls,
        _: CoreSchema,
        handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        return handler(core_schema.str_schema())

    @classmethod
    def __get_validators__(cls) -> Generator[Callable[..., str | None], None, None]:
        if cls.patterns:
            yield cls.validate_pattern
        yield cls.validate

    @classmethod
    def validate(cls, v: str, _: Any = None) -> str:
        if cls.max_length is not None and len(v) > cls.max_length:
            raise ValueError(f"too long to be a valid {cls.__name__}.")
        if cls.min_length is not None and len(v) < cls.min_length:
            raise ValueError(f"too short to be a valid {cls.__name__}.")
        return v

    @classmethod
    def validate_pattern(cls, v: Any, _: Any = None) -> str:
        if not isinstance(v, str):
            raise ValueError(f"{cls.__name__} must be instantiated from a string type.")
        patterns = cls.patterns
        if not isinstance(patterns, list): # pragma: no cover
            patterns = [patterns]
        for pattern in patterns:
            if not pattern.match(v):
                raise ValueError(f"not a valid {cls.__name__}.")
        return v

    def __repr__(self) -> str: # pragma: no cover
        return f'{type(self).__name__}({self})'


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

import pydantic

from canonical.ext.resource import Resource
from canonical.ext.resource import TransientMeta
from .types import Permission


class PermissionQuerySpec(pydantic.BaseModel):
    permissions: set[Permission] = pydantic.Field(
        default=...,
        description=(
            "The set of permissions that are being requested on a "
            "resource."
        )
    )


class PermissionQuery(Resource[TransientMeta], version='permissionqueries.iam.webiam.io/v1'):
    spec: PermissionQuerySpec = pydantic.Field(
        default=...,
        description=(
            "Specifies the properties of the authorization that "
            "the client wants to know."
        )
    )

    @classmethod
    def factory(cls, spec: dict[str, Any]):
        return cls.model_validate({
            'metadata': {},
            'spec': spec
        })


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import datetime
import enum

import canonical
import pydantic
from canonical.ext.resource import Resource


class OrderState(str, enum.Enum):
    placed = 'placed'
    failed = 'failed'
    accepted = 'accepted'
    rejected = 'rejecred'


class FulfillmentStatus(str, enum.Enum):
    observe = 'observe'
    pending = 'pending'
    shipped = 'shipped'


class OrderLineStatus(pydantic.BaseModel):
    line_id: int | None = pydantic.Field(..., alias='lineId')

    fulfillment: FulfillmentStatus

    product_id: int = pydantic.Field(
        default=...,
        alias='productId'
    )

    external_id: int = pydantic.Field(
        default=0,
        alias='externalId'
    )

    class Config:
        allow_population_by_field_name: bool = True


class OrderStatus(pydantic.BaseModel):
    order_id: int | None = pydantic.Field(None, alias='orderId')

    state: OrderState = pydantic.Field(
        default=OrderState.placed,
        title='State',
        description=(
            "Contains the current state of the `Order` resource."
        )
    )

    reason: str | None = pydantic.Field(
        default=None,
        title='Reason',
        description="Explains why the order is in the current state."
    )

    items: list[OrderLineStatus] = []


class OrderLine(pydantic.BaseModel):
    external_product_id: int = pydantic.Field(
        default=...,
        alias='externalProductId'
    )
    sku: str
    amount: int
    external_id: int = pydantic.Field(
        default=0,
        alias='externalId'
    )

    class Config:
        allow_population_by_field_name: bool = True


class ExternalOrderStatus(pydantic.BaseModel):
    pass


class ExternalOrderSpec(pydantic.BaseModel):
    account: str
    service: str
    external_order_id: int = pydantic.Field(
        default=...,
        alias='externalOrder'
    )

    billing_address: canonical.GenericPostalAddress = pydantic.Field(
        default=...,
        alias='billingAddress',
        title='Billing address',
    )

    placed: datetime.datetime = pydantic.Field(
        default=...,
        title='Placed',
        description=(
            "The date/time at which the order was placed in the external "
            "system."
        )
    )

    shipping_address: canonical.GenericPostalAddress | None = pydantic.Field(
        default=None,
        alias='shippingAddress'
    )
    store: str
    items: list[OrderLine] = []

    class Config:
        allow_population_by_field_name: bool = True


class ExternalOrder(Resource, group='uplink.molano.nl', version='v1'):
    spec: ExternalOrderSpec
    status: OrderStatus = pydantic.Field(
        default_factory=OrderStatus
    )

    def generate_name(self) -> str:
        return f'{self.spec.store}/orders/{self.spec.external_order_id}'

    class Config:
        annotations: bool = True
        labels: bool = True


meta = {
    'name': 'molano-uplink.myshopify.com/orders/1',
    'labels': {
        'uplink.molano.nl/platform': 'shopify'
    }
}
spec = {
    'account': 'molano',
    'service': 'molano-uplink',
    'externalOrder': 1,
    'placed': datetime.datetime.utcnow(),
    'store': 'molano-uplink.myshopify.com',
    'billingAddress': {
        'addressee': "Molano B.V.",
        'contact_name': 'Laili Ishaqzai',
        'address1': "Tappersweg 8N",
        'postal_code': '2031ET',
        'city': 'Haarlem',
        'country': 'NL'
    },
    'shippingAddress': {
        'addressee': 'Wizards of Industry Holding B.V.',
        'contact_name': "Cochise Ruhulessin",
        'address1': "Koningin Julianaplein 10",
        'postal_code': "2595AA",
        'city': 'Den Haag',
        'country': 'NL'
    },
    'items': [
        {
            'external_product_id': 1,
            'amount': 2,
            'sku': 'XXXTEST',
            'external_id': 5
        },
        {
            'external_product_id': 1,
            'amount': 2,
            'sku': 'XXXBAR'
        },
    ]
}

dto = ExternalOrder.restore(
    uid=124,
    name='molano-uplink.myshopify.com/orders/1',
    created=datetime.datetime.utcnow(),
    spec=spec,
    metadata=meta
)

#print(dto.json(by_alias=True, exclude_none=True, indent=2))

dto2 = ExternalOrder.new(None, spec)
dto2.annotate('must-refresh', True)
dto2.label('uplink.molano.nl/platform','shopify')
dto2.set_status({
    'items': [
        {
            'line_id': 86868,
            'product_id': 124,
            'fulfillment': FulfillmentStatus.observe,
            'external_id': 5
        },
    ]
})
print(dto2.json(by_alias=True, indent=2))
print(ExternalOrder.Config.pluralname)



# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.

import pydantic


class ListMeta(pydantic.BaseModel):
    """Standard list metadata."""
    model_config = {'populate_by_name': True}

    next_page: str | None = pydantic.Field(
        default=None,
        title='Continue',
        alias='nextPage',
        description=(
            "continue may be set if the user set a limit on the number of items "
            "returned, and indicates that the server has more data available. "
            "The value is opaque and may be used to issue another request to "
            "the endpoint that served this list to retrieve the next set of "
            "available objects. Continuing a consistent list may not be possible "
            "if the server configuration has changed or more than a few minutes "
            "have passed. The resourceVersion field returned when using this "
            "continue value will be identical to the value in the first response, "
            "unless you have received this token from an error message."
        )
    )

    resource_version: str = pydantic.Field(
        default='',
        alias='resourceVersion',
        description=(
            "String that identifies the server's internal version of this object that "
            "can be used by clients to determine when objects have changed. Value "
            "must be treated as opaque by clients and passed unmodified back to the "
            "server. Populated by the system. Read-only."
        )
    )

    @classmethod
    def default(cls):
        return cls()